--  font.e: font library
--  j.babor@gns.cri.nz


--  version 3.10   98-03-08 ----------------------------------------------------
--      each loaded font now has its own private copy of parameters
--      expanded font attributes: bold, shadows, italics, underline & vertical
--      text justification added: left, right and centre as well as
--          multi-line vertical centering - all consolidated in fprint
--      optional extra line and/or paragraph spacing added to font parameters
--      optional extra first line indent added to font parameters
--      procedure freset() added
--      function ffont() expanded
--      font parameters changed
--      defaults changed
--      fgets() function has additional width parameter, scroll, xor cursor 
--          and much better edit facilities
--      texturing with local or global option added to standard font colors
--      wrap & justification defaults automatically restored after each fprint

--      slope printing (temporarily?) abandoned
--      indirect printing (funtions fci() & fsi()) no longer supported


--  font file format -----------------------------------------------------------
--      byte 1: ascii code of the first char      (a1)
--      byte 2: ascii code of the last char       (an)
--      byte 3: width in bytes                    (bw)
--      byte 4: height in pixels                  (h)
--      byte 5: version                           (3)      new!
--      byte 6: baseline height                   (blh)    new!
--      next (an-a1+1) bytes: actual char width   (w)
--      next (an-a1+1) sets of bw*h bytes: simple char bitmaps, left-to-right,
--          top-to-bottom


--  naming conventions ---------------------------------------------------------
--      global constants are capitalized three letter words starting with 'F'
--      global variables are two letter lower case words starting with 'f'
--      global functions & procedures start with 'f'


include graphics.e
include machine.e
include wildcard.e

global constant
    -- some pre-defined constants (anyone silly enough to memorize them?)
    FPA=    -- default set of parameters assigned to each loaded font
       {1,  -- 1: boldface offset
        1,  -- 2: shadow offset: right
        1,  -- 3: shadow offset: down
        0,  -- 4: extra kerning space width
        0,  -- 5: extra underline drop
        1,  -- 6: underline thickness
        1,  -- 7: extra line spacing
        0,  -- 8: extra paragraph spacing
        0,  -- 9: extra first line indent
        0}, --10: tiling (local=0, global=1)

    FDE=-1,     -- detect paper color, used in fgets()

    -- attributes ('additive', e.g.  fa=FBO+FIT  gives you bold italics)
    NUL=0,  -- none: surprise !
    FBO=1,  -- bold
    FSH=2,  -- shadows
    FIT=4,  -- italics
    FUN=8,  -- underline
    FVE=16, -- vertical printing  (not implemented, yet!)

    -- (multi-)line justification
    FNN=0,  -- (or NUL) no justification in either direction,  '\n'
            -- (new line) returns to original fx with line feed
    FLN=1,  -- left justified, no vertical justification
    FCN=2,  -- horizontally centred, no vertical justification
    FRN=3,  -- right justified, no vertical justification
    -- vertically centred lines (about initial fy)
    FLC=4,  -- left justified, vertically centred
    FCC=5,  -- centred in both directions
    FRC=6,  -- right justified, vertically centred

    -- line wrap
    FNW=0,  -- (or NUL) no wrap
    FCW=1,  -- char based wrap
    FWW=FCW -- word based wrap (still defaults to char wrap (again,
            -- sorry, not yet implemented, perhaps next time!)

global object fc    -- text color or tile

sequence
    names,          -- font filenames
    fonts,          -- char images
    bwidths,        -- widths (bytes)
    heights,        -- heights (pixels)
    blhs,           -- baseline heights (pixels)
    widths,         -- sequences of char widths (pixels)
    first_chars,    -- first chars
    last_chars,     -- last chars
    versions,       -- main font version (>3.00 only)
    params,         -- sequence of all font parameters
    io              -- italics offset

global sequence
    fm,             -- current margins
    fp              -- current font parameters

integer
    bo,             -- boldface offset
    f,              -- current font
    version

global integer
    fx,             -- text pointer: left edge char space
    fy,             -- text pointer: top  edge char space
    fb,             -- background (paper) color
    fs,             -- shadow color
    fa,             -- current combined attributes
    fj,             -- text justification (see global constants above)
    fw              -- line wrap (see global constants above)


global function reverse(sequence s)
    -- return sequence with elements in reverse order
    -- now global to make it available to widgets
    sequence rs
    rs={}
    for i=1 to length(s) do rs=prepend(rs,s[i]) end for
    return rs end function

procedure xorpixel(object c, sequence xy)
    -- pixel(s) xor-ed with background, same syntax as pixel()
    if sequence(c) then pixel(xor_bits(get_pixel(xy&length(c)),c),xy)
    else pixel(xor_bits(get_pixel(xy),c),xy) end if end procedure

procedure xorbox(integer c,sequence p1,sequence p2)
    -- draw rectangle of color c xor-ed with background,
    -- p1 & p2 are diagonal points
    for y=p1[2] to p2[2] do
        xorpixel(repeat(c,p2[1]-p1[1]+1),{p1[1],y}) end for end procedure

function loaded(sequence name)
    -- returns previously allocated handle if font already loaded, else zero
    integer i,f
    f=0   i=0
    while i<length(names) do
        i=i+1
        if compare(name,names[i])=0 then
            f=i
            exit end if end while
    return f end function

global procedure freset()
    -- reset entire environment of current font to defaults
    fx=0        -- left edge of char space
    fy=0        -- top edge of char space
    fa=0        -- font attribute (none)
    fb=FDE      -- detect background (paper) color
    fc=WHITE    -- text color
    fs=GRAY     -- shadow color
    fj=FNN      -- no line justification
    fm={0,1024} -- left, right margin
    fw=FNW      -- no line wrap
    fp=FPA      -- default parameters
end procedure   -- freset

global procedure fselect(integer h)
    -- makes font with handle h current
    if h>0 and h<=length(fonts) then
        -- save current parameters if font previously loaded
        if f then params[f]=fp end if
        f=h
        fp=params[f]  -- retrieve saved parameters
        -- set italics offsets
        io={}
        for i=1 to heights[f] do io=io&floor((blhs[f]-i)/4) end for
    end if end procedure  -- fselect

global function fload(sequence file_name)
    -- load font file into memory and make available for selection
    integer blh,n,fn,h,bw,fc,lc,np,len
    sequence bits,font,cs,w
    -- if font is already loaded, do not do it again, just select it
    file_name=lower(file_name)
    n=loaded(file_name)
    if n then fselect(n)
    else
        fn=open(file_name, "rb")
        if fn=-1 then
            puts(1, "Font load error: " & file_name & " not found !\n")
            abort(1) end if
        names=append(names,file_name)
        fc=getc(fn)
        lc=getc(fn)
        bw=getc(fn)
        h=getc(fn)
        version=getc(fn)
        blh=getc(fn)                  -- baseline height
        if blh=0 then blh=h end if    -- pre version 3 fonts
        w={}
        font={}
        for i=fc to lc do w=w & getc(fn) end for
        for i=fc to lc do
            cs={}
            for row=1 to h do
                bits={}
                for col=1 to bw do bits=int_to_bits(getc(fn),8) & bits end for
                bits=reverse(bits)
                np=0  len=length(bits)
                for j=1 to len do
                   if bits[j] then np=np+1
                   elsif np then
                       cs=append(cs,{j-np-1,row-1,np})
                       np=0 end if end for
                if np then cs=append(cs,{len-np,row-1,np}) end if end for
            font=append(font, cs) end for
        close(fn)
        fonts=append(fonts, font)
        f=length(fonts)   -- make the last loaded font current
        first_chars=append(first_chars,fc)
        last_chars=append(last_chars,lc)
        widths=append(widths, w)
        bwidths=bwidths & bw
        heights=heights & h
        blhs=blhs & blh
        versions=versions & version
        params=append(params,FPA)
        -- set italics offsets for new font
        io={}
        for i=1 to heights[f] do
            io=io & floor((blhs[f]-i)/4) end for end if
    return f end function  -- fload

global function ffont(integer f)
    -- f is handle of currently loaded font
    -- return a sequence of font particulars
    return  {names[f],          -- font filename
             versions[f],       -- only from version 3.0
             first_chars[f],    -- first char ASCII code
             last_chars[f],     -- last char ASCII code
             bwidths[f],        -- char width in bytes
             heights[f],        -- nominal char height in pixels
             blhs[f]}           -- baseline height (v 3.0+)
end function

global function fheight(integer f)
    -- return nominal height of a currently loaded font with handle f
    return heights[f] end function

global function flength(sequence text)
    -- return sequence of lengths (in pixels) of string text
    object len
    integer b,l
    len={}   l=0
    b=and_bits(FBO,fa)*fp[1]               -- current bold offset
    for i=1 to length(text) do
        if text[i]<=0 then
            b=and_bits(FBO,text[i])*fp[1]    -- new bold offset
        elsif text[i]=10 then                 -- new line
            len=len&l   l=0
        elsif text[i]>=first_chars[f] and text[i]<=last_chars[f] then
            l=l+widths[f][text[i]-first_chars[f]+1]+b+fp[4] end if end for
    return len&l end function   -- flength

global procedure fchar(integer ch)
    -- prints character ch at current text pointer location (fx,fy) in
    -- currently selected font - not recommended for direct use
    integer c,dx,n,mx,my,nx,ny
    sequence cs,ps,ss
    if ch>=first_chars[f] and ch<=last_chars[f] then      -- printable char
        c=ch-first_chars[f]+1
        ps=fonts[f][c]
        dx=widths[f][c]+bo              -- char space width
        if integer(fc) then             -- single color font
            if and_bits(FIT,fa) then
                if and_bits(FSH,fa) and (fp[2] or fp[3]) then    -- shadows
                    cs=repeat(fs,dx)
                    for i=1 to length(ps) do
                        pixel(cs[1..ps[i][3]+bo],ps[i][1..2]+{fx+fp[2]+
                        io[ps[i][2]+1],fy+fp[3]}) end for
                end if
                cs=repeat(fc,dx)
                for i=1 to length(ps) do
                    pixel(cs[1..ps[i][3]+bo],ps[i][1..2]+
                    {fx+io[ps[i][2]+1],fy}) end for
            else
                if and_bits(FSH,fa) and (fp[2] or fp[3]) then   -- shadows
                    cs=repeat(fs,dx)
                    for i=1 to length(ps) do pixel(cs[1..ps[i][3]+
                        bo],ps[i][1..2]+{fx+fp[2],fy+fp[3]}) end for end if
                cs=repeat(fc,dx)
                for i=1 to length(ps) do
                    pixel(cs[1..ps[i][3]+bo],ps[i][1..2]+{fx,fy}) end for
            end if
            if and_bits(FUN,fa) and fp[6] then
                if and_bits(FSH,fa) and (fp[2] or fp[3]) then
                    cs=repeat(fs,dx+fp[4])
                    for i=0 to fp[6]-1 do pixel(cs,{fx+fp[2],fy+fp[3]+fp[5]+heights[f]+i}) end for
                end if
                cs=repeat(fc,dx+fp[4])
                for i=0 to fp[6]-1 do pixel(cs,{fx,fy+fp[5]+heights[f]+i}) end for
            end if
        else        -- tiled char
            nx=length(fc[1])
            ny=length(fc)
            if and_bits(FIT,fa) then
                if and_bits(FSH,fa) and (fp[2] or fp[3]) then    -- shadows
                    cs=repeat(fs,dx)
                    for i=1 to length(ps) do
                        pixel(cs[1..ps[i][3]+bo],ps[i][1..2]+{fx+fp[2]+io[ps[i][2]+1],fy+fp[3]}) end for
                end if
                n=floor((dx+io[1])/nx)
                cs=fc
                for i=1 to ny do
                    for j=1 to n do cs[i]=cs[i]&fc[i] end for end for
                if fp[10] then  -- global tile alignment
                    mx=1+remainder(fx,nx)
                    for i=1 to ny do
                        cs[i]=fc[i][mx..nx] & cs[i] end for end if
                for i=1 to length(ps) do
                    if fp[10] then  -- global tile alignment
                        my=1+remainder(fy+ps[i][2],ny)
                    else            -- local tile alignment
                        my=1+remainder(ps[i][2],ny)
                    end if
                    pixel(cs[my][ps[i][1]+io[ps[i][2]+1]+1..ps[i][1]+ps[i][3]+
                          io[ps[i][2]+1]+bo],ps[i][1..2]+{fx+io[ps[i][2]+1],fy})
                end for
            else
                if and_bits(FSH,fa) and (fp[2] or fp[3]) then   -- shadows
                    cs=repeat(fs,dx)
                    for i=1 to length(ps) do
                        pixel(cs[1..ps[i][3]+bo],ps[i][1..2]+{fx+fp[2],fy+fp[3]}) end for
                end if
                n=floor(dx/nx)
                cs=fc
                for i=1 to ny do
                    for j=1 to n do cs[i]=cs[i]&fc[i] end for end for
                if fp[10] then      -- global tile alignment
                    mx=1+remainder(fx,nx)
                    for i=1 to ny do
                        cs[i]=fc[i][mx..nx] & cs[i] end for end if
                for i=1 to length(ps) do
                    if fp[10] then  -- global tile alignment
                        my=1+remainder(fy+ps[i][2],ny)
                    else            -- local tile alignment
                        my=1+remainder(ps[i][2],ny)
                    end if
                    pixel(cs[my][ps[i][1]+1..ps[i][1]+ps[i][3]+bo],
                          ps[i][1..2]+{fx,fy}) end for
            end if
            if and_bits(FUN,fa) and fp[6] then
                if and_bits(FSH,fa) and (fp[2] or fp[3]) then
                    ss=repeat(fs,dx+fp[4])
                    for i=0 to fp[6]-1 do pixel(ss,{fx+fp[2],fy+fp[3]+fp[5]+heights[f]+i}) end for
                end if
                for i=0 to fp[6]-1 do
                    if fp[8] then  -- global tile alignment
                        my=1+remainder(fy+i+fp[5]+heights[f],ny)
                    else            -- local tile alignment
                        my=1+remainder(i+fp[5]+heights[f],ny)
                    end if
                    pixel(cs[my][1..dx+fp[4]],{fx,fy+fp[5]+heights[f]+i})
                end for
            end if
        end if
        fx=fx+dx+fp[4]  -- adjust text pointer
    end if
end procedure  -- fchar

global procedure fprint(sequence s)
    -- print string s in currently selected font, using current margins
    -- and current justification
    -- return *automatically* to sensible justification & wrap defaults
    integer c,j,n,x
    sequence flen
    bo=and_bits(FBO,fa)*fp[1]                   -- new bold actual offset
    if fj=FNN then                              -- no justification
        x=fx
        for i=1 to length(s) do
            c=s[i]
            if c>255 then fc=c-256              -- new text color
            elsif c=10 then                     -- new line
                fx=x                            -- starting fx
                fy=fy+heights[f]+fp[7]+fp[8]    -- line feed
            elsif c<=0 then                     -- reset attributes
                fa=-c
                bo=and_bits(FBO,fa)*fp[1]       -- new bold offset
            else fchar(c) end if end for        -- print if printable
    elsif fj=FLN then                           -- left justified
        for i=1 to length(s) do
            c=s[i]
            if c>255 then fc=c-256              -- new text color
            elsif c=10 then                     -- new line
                fx=fm[1]+fp[9]                  -- left margin
                fy=fy+heights[f]+fp[7]+fp[8]
            elsif c<=0 then                     -- reset attributes
                fa=-c
                bo=and_bits(FBO,fa)*fp[1]       -- new bold offset
            elsif fw=FCW and c>=first_chars[f] and c<=last_chars[f] then
                if fx+widths[f][c-first_chars[f]+1]+bo>fm[2] then
                    fx=fm[1]
                    fy=fy+heights[f]+fp[7] end if
                fchar(c)
            else fchar(c) end if end for        -- no wrap
    elsif fj=FCN then                           -- horizontally centred
        x=fx    j=1
        flen=flength(s)
        fx=floor(x-(flen[j]+1)/2)
        for i=1 to length(s) do
            c=s[i]
            if c>255 then fc=c-256              -- new text color
            elsif c=10 then                     -- new line
                j=j+1
                fx=floor(x-(flen[j]+1)/2)
                fy=fy+heights[f]+fp[7]+fp[8]
            elsif c<=0 then                     -- reset attributes
                fa=-c
                bo=and_bits(FBO,fa)*fp[1]       -- new bold offset
            else fchar(c) end if
        end for
    elsif fj=FRN then                           -- right justified
        x=fx    j=1
        flen=flength(s)
        fx=x-flen[j]
        for i=1 to length(s) do
            c=s[i]
            if c>255 then fc=c-256              -- new text color
            elsif c=10 then                     -- new line
                j=j+1
                fx=x-flen[j]
                fy=fy+heights[f]+fp[7]+fp[8]
            elsif c<=0 then                     -- reset attributes
                fa=-c
                bo=and_bits(FBO,fa)*fp[1]       -- new bold offset
            else fchar(c) end if end for
    elsif fj=FLC then                           -- left, vertically centred
        n=length(flength(s))
        fy=floor(fy-(n*heights[f]+(n-1)*(fp[7]+fp[8])+1)/2)
        for i=1 to length(s) do
            c=s[i]
            if c>255 then fc=c-256              -- new text color
            elsif c=10 then                     -- new line
                fx=fm[1]                        -- left_margin
                fy=fy+heights[f]+fp[7]+fp[8]
            elsif c<=0 then                     -- reset attributes
                fa=-c
                bo=and_bits(FBO,fa)*fp[1]       -- new bold offset
            else fchar(c) end if end for
    elsif fj=FCC then   -- text horizontally & vertically centred
        x=fx    j=1
        flen=flength(s)
        fx=floor(x-(flen[j]+1)/2)
        n=length(flen)
        fy=floor(fy-(n*heights[f]+(n-1)*(fp[7]+fp[8])+1)/2)
        for i=1 to length(s) do
            c=s[i]
            if c>255 then fc=c-256              -- new text color
            elsif c=10 then                     -- new line
                j=j+1
                fx=floor(x-(flen[j]+1)/2)
                fy=fy+heights[f]+fp[7]+fp[8]
            elsif c<=0 then                     -- reset attributes
                fa=-c
                bo=and_bits(FBO,fa)*fp[1]       -- new bold offset
            else fchar(c) end if
        end for
    else    -- FRC, text right justified, vertically centred
        x=fx    j=1
        flen=flength(s)
        fx=x-flen[j]
        n=length(flen)
        fy=floor(fy-(n*heights[f]+(n-1)*(fp[7]+fp[8])+1)/2)
        for i=1 to length(s) do
            c=s[i]
            if c>255 then fc=c-256              -- new text color
            elsif c=10 then                     -- new line
                j=j+1
                fx=x-flen[j]
                fy=fy+heights[f]+fp[7]+fp[8]
            elsif c<=0 then                     -- reset attributes
                fa=-c
                bo=and_bits(FBO,fa)*fp[1]       -- new bold offset
            else fchar(c) end if
        end for
    end if
    fj=FNN  fw=FNW      -- automatic return to sensible defaults
end procedure  -- fprint

function strip(sequence s)
    -- return string s stripped of all formatting characters & \n,\r, \t
    sequence ss
    integer c
    ss={}
    for i=1 to length(s) do
        c=s[i]
        if c!='\n' and c!='\r' and c!='\t' and c>=first_chars[f]
            and c<=last_chars[f] then ss=ss&c end if  end for
    return ss end function

function flen(sequence text)
    -- simplified version of flength()
    -- return length of already stripped text, in pixels
    integer bo,len
    bo=and_bits(FBO,fa)*fp[1]   -- new bold actual offset
    len=0
    for i=1 to length(text) do
        len=len+widths[f][text[i]-first_chars[f]+1]+bo+fp[4] end for
    return len end function

global function fgets(sequence s,integer w)
    --  prompted screen input of a string with primitive editing
    --  all formatting characters in prompt string s are ignored (stripped)
    --  attributes other than BOLD are, temporally, ignored
    --  text justification is set, temporally, to FNN
    atom t
    integer
        att,jus,temp,
        p,      -- position, general
        cc,     -- cursor color
        key,dx,dy,
        pc,     -- paper color
        flag,   -- cursor flag, drawn/erased
        len,    -- length, in pixels
        l,      -- current string length, number of chars
        ci,     -- cursor index, index of char under it
        fvc,    -- index of first visible char
        sx      -- starting fx, left edge of box

    att=fa
    fa=and_bits(fa,FBO)  -- allow bold mode only
    jus=fj
    fj=FLN
    s=strip(s)
    dy=heights[f]-1
    if fb=FDE then pc=get_pixel({fx,fy}) else pc=fb end if
    sx=fx
    l=length(s)
    fvc=1
    ci=l+1
    while flen(s[fvc..l])>w do fvc=fvc+1 end while
    fprint(s[fvc..l])

    -- blinking xor cursor
    cc=BLUE
    while 1 do
        t=time()
        xorbox(cc,{fx,fy},{fx+1,fy+heights[f]-1})
        flag=1                      -- cursor drawn
        while 1 do
            if time()-t > 0.2 then
                t=time()
                xorbox(cc,{fx,fy},{fx+1,fy+heights[f]-1})
                flag=not flag  end if
            key=get_key()
            if key!=-1 then
                -- erase current cursor if necessary
                if flag then
                    xorbox(cc,{fx,fy},{fx+1,fy+heights[f]-1})
                end if
                exit end if end while
        if key=331 then      -- left arrow
            if ci>1 then
                ci=ci-1
                if ci<fvc then  -- at left edge of box - scroll
                    fvc=fvc-1
                    len=flen(s[ci..l])
                    if len>w then
                        p=l
                        while len>w do
                            p=p-1
                            len=flen(s[ci..p]) end while
                        polygon(pc,1,{{sx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{sx,fy+dy}})
                        fx=sx
                        fprint(s[ci..p])
                        fx=sx
                    else
                        polygon(pc,1,{{sx,fy},{sx+len-1,fy},{sx+len-1,fy+dy},{sx,fy+dy}})
                        fx=sx
                        fprint(s[ci..l])
                        fx=sx end if
                else fx=fx-flen({s[ci]}) end if end if
        elsif key=333 then      -- right arrow
            if ci<=l then
                len=flen(s[fvc..ci])
                if len>w then   -- right box edge
                    fvc=fvc+1
                    len=flen(s[fvc..ci])
                    while len>w do
                        fvc=fvc+1
                        len=flen(s[fvc..ci]) end while
                    fx=sx
                    polygon(pc,1,{{fx,fy},{fx+w-1,fy},{fx+w-1,fy+dy},{fx,fy+dy}})
                    fprint(s[fvc..ci])
                    if ci<l then
                        if len+flen({s[ci+1]})<=w then
                            fprint({s[ci+1]}) end if end if
                    fx=sx+len
                else fx=fx+flen({s[ci]}) end if
                ci=ci+1 end if
        elsif key=327 then  -- home key
            if fvc>1 then
                p=0
                while p<=l do
                    len=flen(s[1..p])
                    if flen(s[1..p])>w then
                        p=p-1
                        exit
                    else p=p+1 end if end while
                polygon(pc,1,{{sx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{sx,fy+dy}})
                fx=sx
                fprint(s[1..p]) end if
            ci=1
            fx=sx
            fvc=1
        elsif key=335 then  -- end key
            if ci!=l+1 then
                fvc=1
                ci=l+1
                while flen(s[fvc..l])>w do fvc=fvc+1 end while
                polygon(pc,1,{{sx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{sx,fy+dy}})
                fx=sx
                fprint(s[fvc..l]) end if
        elsif key=339 then  -- delete
            if ci and ci<=l then
                temp=fx
                len=flen(s[ci+1..l])
                if len>w-fx+sx then
                    p=l
                    while len>w-fx+sx do
                        p=p-1
                        len=flen(s[ci+1..p]) end while
                    polygon(pc,1,{{fx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{fx,fy+dy}})
                    fprint(s[ci+1..p])
                else
                    len=len+flen({s[ci]})
                    if len>w-fx+sx then len=w-fx+sx end if
                    polygon(pc,1,{{fx,fy},{fx+len-1,fy},{fx+len-1,fy+dy},{fx,fy+dy}})
                    fprint(s[ci+1..l]) end if
                fx=temp
                s=s[1..ci-1]&s[ci+1..l]
                l=l-1 end if
        elsif key=8 then    -- backspace
            if ci>1 then
                ci=ci-1
                l=l-1
                if ci>l then    -- at end of string
                    if ci<fvc then  -- at left edge of box
                        fvc=fvc-1
                    else
                        dx=flen({s[ci]})
                        fx=fx-dx
                        -- and erase the last char
                        polygon(pc,1,{{fx,fy},{fx+dx-1,fy},{fx+dx-1,fy+dy},
                        {fx,fy+dy}}) end if
                    s=s[1..l]
                else    -- inside string
                    if ci<fvc then fvc=fvc-1    -- at left edge of box
                    else
                        dx=flen({s[ci]})
                        fx=fx-dx
                        temp=fx
                        len=flen(s[ci+1..l+1])
                        if len>w-fx+sx then
                            p=l
                            len=flen(s[ci+1..p])
                            while len>w-fx+sx do
                                p=p-1
                                len=flen(s[ci+1..p]) end while
                            polygon(pc,1,{{fx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{fx,fy+dy}})
                            fprint(s[ci+1..p])
                            fx=temp
                        else
                            polygon(pc,1,{{fx,fy},{fx+len+dx-1,fy},{fx+len+dx-1,fy+dy},{fx,fy+dy}})
                            fprint(s[ci+1..l+1])
                            fx=temp end if end if
                    s=s[1..ci-1]&s[ci+1..l+1] end if end if
        elsif key=13 then exit  -- enter
        elsif key=27 then       -- escape
            s=""
            exit
        elsif key>0 and key>=first_chars[f] and key<=last_chars[f] then
            if ci>l then        -- at end of string
                s=s&key
                l=l+1
                ci=ci+1
                len=flen(s[fvc..l])
                if len<=w then fchar(key)
                else            -- past right edge of box - scroll left
                    while len>w do
                        fvc=fvc+1
                        len=flen(s[fvc..l]) end while
                    polygon(pc,1,{{sx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{sx,fy+dy}})
                    fx=sx
                    fprint(s[fvc..l]) end if
            else
                s=s[1..ci-1]&key&s[ci..l]
                l=l+1
                temp=fx
                len=flen(s[fvc..l])
                if len<=w then
                    polygon(pc,1,{{fx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{fx,fy+dy}})
                    fprint(s[ci..l])
                    fx=temp+flen({s[ci]})
                else            -- scroll right end further right
                    p=l
                    while len>w do
                        p=p-1
                        len=flen(s[fvc..p]) end while
                    polygon(pc,1,{{sx,fy},{sx+w-1,fy},{sx+w-1,fy+dy},{sx,fy+dy}})
                    fx=sx
                    fprint(s[fvc..p])
                    fx=temp+flen({s[ci]}) end if
                ci=ci+1 end if end if end while
    fj=jus    fa=att
    return s end function   -- fgets

-- init variables
f=0
fonts={}
names={}
versions={}
first_chars={}
last_chars={}
bwidths={}
heights={}
widths={}
blhs={}
params={}
freset()
